import json

from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.template.defaultfilters import truncatechars
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import localtime

from generic_m2m_field.models import MultipleDBGenericManyToManyField

from chamber.models import SmartModel

from enumfields import NumEnumField

from .enums import InputLoggedRequestType, LoggedRequestStatus, CeleryTaskInvocationLogState, CeleryTaskRunLogState


def display_json(value, indent=4):
    return json.dumps(value, indent=indent, ensure_ascii=False, cls=DjangoJSONEncoder)


class LoggedRequest(SmartModel):

    host = models.CharField(_('host'), max_length=255, null=False, blank=False, db_index=True)
    method = models.SlugField(_('method'), max_length=7, null=False, blank=False, db_index=True)
    path = models.CharField(_('URL path'), max_length=2000, null=False, blank=True, db_index=True)
    queries = models.JSONField(_('queries'), null=True, blank=True, encoder=DjangoJSONEncoder)
    is_secure = models.BooleanField(_('HTTPS connection'), default=False, null=False, blank=False)
    slug = models.CharField(_('slug'), null=True, blank=True, db_index=True, max_length=255)

    # Request information
    request_timestamp = models.DateTimeField(_('request timestamp'), null=False, blank=False, db_index=True)
    request_headers = models.JSONField(_('request headers'), null=True, blank=True, encoder=DjangoJSONEncoder)
    request_body = models.TextField(_('request body'), null=False, blank=True)

    # Response information
    response_timestamp = models.DateTimeField(_('response timestamp'), null=True, blank=True, db_index=True)
    response_code = models.PositiveSmallIntegerField(_('response code'), null=True, blank=True)
    response_headers = models.JSONField(_('response headers'), null=True, blank=True, encoder=DjangoJSONEncoder)
    response_body = models.TextField(_('response body'), null=True, blank=True)
    response_time = models.FloatField(_('response time'), null=True, blank=True)

    status = NumEnumField(verbose_name=_('status'), enum=LoggedRequestStatus, null=False, blank=False,
                          default=LoggedRequestStatus.INCOMPLETE)
    error_description = models.TextField(_('error description'), null=True, blank=True)
    exception_name = models.CharField(_('exception name'), null=True, blank=True, max_length=255)

    def short_queries(self):
        return truncatechars(display_json(self.queries, indent=0), 50)
    short_queries.short_description = _('queries')
    short_queries.filter_by = 'queries'

    def short_request_headers(self):
        return truncatechars(display_json(self.request_headers, indent=0), 50)
    short_request_headers.short_description = _('request headers')
    short_request_headers.filter_by = 'request_headers'

    def short_path(self):
        return truncatechars(self.path, 50)
    short_path.short_description = _('Path')
    short_path.filter_by = 'path'
    short_path.order_by = 'path'

    def short_response_body(self):
        return truncatechars(self.response_body, 50) if self.response_body is not None else None
    short_response_body.short_description = _('response body')
    short_response_body.filter_by = 'response_body'

    def short_request_body(self):
        return truncatechars(self.request_body, 50)
    short_request_body.short_description = _('request body')
    short_request_body.filter_by = 'request_body'

    def __str__(self):
        return ' '.join(
            (force_text(v) for v in (
                self.slug, self.response_code, localtime(self.request_timestamp.replace(microsecond=0))
            ) if v)
        )

    class Meta:
        abstract = True

    class UIMeta:
        default_ui_filter_by = 'id'


class InputLoggedRequest(LoggedRequest):

    user_id = models.TextField(verbose_name=_('user ID'), null=True, blank=True, db_index=True)
    ip = models.GenericIPAddressField(_('IP address'), null=False, blank=False, db_index=True)
    type = NumEnumField(verbose_name=_('type'), enum=InputLoggedRequestType,
                        default=InputLoggedRequestType.COMMON_REQUEST,
                        null=False, blank=False, db_index=True)
    related_objects = MultipleDBGenericManyToManyField()

    class Meta:
        verbose_name = _('input logged request')
        verbose_name_plural = _('input logged requests')
        ordering = ('-created_at',)

    class SmartMeta:
        is_cleaned_pre_save = False

    @cached_property
    def user(self):
        return get_user_model().objects.filter(pk=self.user_id).first()


class OutputLoggedRequest(LoggedRequest):

    related_objects = MultipleDBGenericManyToManyField()

    class Meta:
        verbose_name = _('output logged request')
        verbose_name_plural = _('output logged requests')
        ordering = ('-created_at',)

    class SmartMeta:
        is_cleaned_pre_save = False


class CommandLog(SmartModel):
    """
    Represents a log of a command run.

    Attributes:
        start: Date and time when command was started.
        stop: Date and time when command finished.
        time: Command processing time in miliseconds.
        name: Name of the command.
        options: Arguments/options the command was run with.
        executed_from_command_line: Flag that indicates if command was run from the command line.
        output: Standard and error output of the command.
        is_successful: Flag that indicates if command finished successfully.
    """
    start = models.DateTimeField(verbose_name=_('start'), blank=False, null=False, editable=False)
    stop = models.DateTimeField(verbose_name=_('stop'), blank=True, null=True, editable=False)
    time = models.FloatField(verbose_name=_('time'), null=True, blank=True)
    name = models.CharField(max_length=250, blank=False, null=False, editable=False, db_index=True,
                            verbose_name=_('name'))
    input = models.TextField(verbose_name=_('input'), blank=False, null=False, editable=False)
    executed_from_command_line = models.BooleanField(verbose_name=_('executed from command line'),
                                                     blank=False, null=False, default=False, editable=False)
    output = models.TextField(verbose_name=_('output'), blank=True, null=True, editable=False)
    is_successful = models.BooleanField(verbose_name=_('finished successfully'),
                                        blank=False, null=False, default=False, editable=False)
    error_message = models.TextField(verbose_name=_('error message'), null=True, blank=True, editable=False)
    related_objects = MultipleDBGenericManyToManyField()

    class Meta:
        verbose_name = _('command log')
        verbose_name_plural = _('command logs')
        ordering = ('-created_at',)

    class UIMeta:
        default_ui_filter_by = 'id'


class CeleryTaskInvocationLogManager(models.Manager):

    def filter_not_started(self):
        return self.filter(
            state__in={
                CeleryTaskInvocationLogState.WAITING,
                CeleryTaskInvocationLogState.TRIGGERED
            }
        )

    def filter_processing(self, related_objects=None, **kwargs):
        qs = self.filter(
            state__in={
                CeleryTaskInvocationLogState.ACTIVE,
                CeleryTaskInvocationLogState.WAITING,
                CeleryTaskInvocationLogState.TRIGGERED
            }
        ).filter(**kwargs)

        if related_objects:
            for related_object in related_objects:
                qs = qs.filter(
                    related_objects__object_id=related_object.pk,
                    related_objects__object_ct_id=ContentType.objects.get_for_model(related_object).pk
                )
        return qs


class CeleryTaskInvocationLog(SmartModel):

    invocation_id = models.CharField(
        verbose_name=_('invocation ID'),
        max_length=250,
        db_index=True,
        unique=True
    )
    celery_task_id = models.CharField(
        verbose_name=_('celery ID'),
        null=True,
        blank=True,
        max_length=250,
        db_index=True
    )
    name = models.CharField(
        verbose_name=_('task name'),
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    queue_name = models.CharField(
        verbose_name=_('queue name'),
        null=True,
        blank=True,
        max_length=250
    )
    applied_at = models.DateTimeField(
        verbose_name=_('applied at'),
        null=False,
        blank=False
    )
    triggered_at = models.DateTimeField(
        verbose_name=_('triggered at'),
        null=True,
        blank=True
    )
    is_unique = models.BooleanField(
        verbose_name=_('is unique')
    )
    is_async = models.BooleanField(
        verbose_name=_('is async')
    )
    is_duplicate = models.BooleanField(
        verbose_name=_('is duplicate')
    )
    is_on_commit = models.BooleanField(
        verbose_name=_('is on commit')
    )
    input = models.TextField(
        verbose_name=_('input'),
        blank=True,
        null=True,
        editable=False
    )
    task_args = models.JSONField(
        verbose_name=_('task args'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    task_kwargs = models.JSONField(
        verbose_name=_('task kwargs'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    estimated_time_of_first_arrival = models.DateTimeField(
        verbose_name=_('estimated time of first arrival'),
        null=True,
        blank=True
    )
    expires_at = models.DateTimeField(
        verbose_name=_('time of expiration'),
        null=True,
        blank=True
    )
    stale_at = models.DateTimeField(
        verbose_name=_('stale task time'),
        null=True,
        blank=True
    )
    state = NumEnumField(
        verbose_name=_('state'),
        null=False,
        blank=False,
        enum=CeleryTaskInvocationLogState,
        default=CeleryTaskInvocationLogState.WAITING,
        db_index=True
    )

    related_objects = MultipleDBGenericManyToManyField()

    class Meta:
        verbose_name = _('celery task')
        verbose_name_plural = _('celery tasks')
        ordering = ('-created_at',)

    class UIMeta:
        default_ui_filter_by = 'id'

    def __str__(self):
        return '{} ({})'.format(self.name, self.state.label)

    @property
    def runs(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id)

    @property
    def last_run(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id).last()

    @property
    def first_run(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id).first()

    def get_start(self):
        first_run = self.first_run
        return first_run.start if first_run else None
    get_start.short_description = _('start')

    def get_stop(self):
        last_run = self.last_run
        return last_run.stop if last_run else None
    get_stop.short_description = _('stop')


class CeleryTaskRunLog(SmartModel):

    celery_task_id = models.CharField(
        verbose_name=_('celery ID'),
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    start = models.DateTimeField(
        verbose_name=_('start'),
        null=True,
        blank=True
    )
    stop = models.DateTimeField(
        verbose_name=_('stop'),
        null=True,
        blank=True
    )
    time = models.FloatField(
        verbose_name=_('time'),
        null=True,
        blank=True
    )
    state = NumEnumField(
        verbose_name=_('state'),
        null=False,
        blank=False,
        enum=CeleryTaskRunLogState,
        default=CeleryTaskRunLogState.ACTIVE,
        db_index=True
    )
    name = models.CharField(
        verbose_name=_('task name'),
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    task_args = models.JSONField(
        verbose_name=_('task args'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    task_kwargs = models.JSONField(
        verbose_name=_('task kwargs'),
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    result = models.JSONField(
        verbose_name=_('result'),
        blank=True,
        null=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    error_message = models.TextField(
        verbose_name=_('error message'),
        null=True,
        blank=True
    )
    output = models.TextField(
        verbose_name=_('output'),
        blank=True,
        null=True,
        editable=False
    )
    retries = models.PositiveSmallIntegerField(
        verbose_name=_('retries'),
        null=False,
        blank=False,
        default=0
    )
    estimated_time_of_next_retry = models.DateTimeField(
        verbose_name=_('estimated time of arrival'),
        null=True,
        blank=True
    )
    queue_name = models.CharField(
        verbose_name=_('queue name'),
        null=True,
        blank=True,
        max_length=250
    )
    related_objects = MultipleDBGenericManyToManyField()

    class Meta:
        verbose_name = _('celery task run')
        verbose_name_plural = _('celery tasks run')
        ordering = ('created_at',)

    def __str__(self):
        return '{} ({})'.format(self.name, self.get_state_display())

    def get_task_invocation_logs(self):
        return CeleryTaskInvocationLog.objects.filter(celery_task_id=self.celery_task_id)



import json

from django.utils.timezone import now

from celery import current_app
from celery.exceptions import NotRegistered

from elasticsearch import ConflictError
from elasticsearch_dsl import Q

from security.config import settings
from security.enums import (
    RequestLogState, CeleryTaskInvocationLogState, CeleryTaskRunLogState, CommandState
)
from security.backends.writer import BaseBackendWriter

from .models import (
    InputRequestLog, OutputRequestLog, CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog, get_key_from_object,
    get_response_state, get_log_model_from_logger_name
)


class ElasticsearchBackendWriter(BaseBackendWriter):

    def input_request_started(self, logger):
        related_objects = [
            get_key_from_object(related_object) for related_object in logger.related_objects
        ]

        input_request_log = InputRequestLog(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=RequestLogState.INCOMPLETE,
            related_objects=related_objects,
            **logger.data
        )
        input_request_log.meta.id = logger.id
        if logger.parent_with_id:
            input_request_log.parent_log = '{}|{}'.format(logger.parent_with_id.name, logger.parent_with_id.id)

        logger.backend_logs.elasticsearch = input_request_log
        input_request_log.save()

    def input_request_finished(self, logger):
        input_request_log = logger.backend_logs.elasticsearch
        input_request_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=get_response_state(logger.data['response_code']),
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def input_request_error(self, logger):
        input_request_log = logger.backend_logs.elasticsearch
        input_request_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def output_request_started(self, logger):
        related_objects = [
            get_key_from_object(related_object) for related_object in logger.related_objects
        ]
        output_request_log = OutputRequestLog(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=RequestLogState.INCOMPLETE,
            related_objects=related_objects,
            update_only_changed_fields=True,
            **logger.data
        )
        output_request_log.meta.id = logger.id
        if logger.parent_with_id:
            output_request_log.parent_log = '{}|{}'.format(logger.parent_with_id.name, logger.parent_with_id.id)
        logger.backend_logs.elasticsearch = output_request_log
        output_request_log.save()

    def output_request_finished(self, logger):
        output_request_log = logger.backend_logs.elasticsearch
        output_request_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=get_response_state(logger.data['response_code']),
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def output_request_error(self, logger):
        output_request_log = logger.backend_logs.elasticsearch
        output_request_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=RequestLogState.ERROR,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def command_started(self, logger):
        related_objects = [
            get_key_from_object(related_object) for related_object in logger.related_objects
        ]
        command_log = CommandLog(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CommandState.ACTIVE,
            related_objects=related_objects,
            **logger.data
        )
        command_log.meta.id = logger.id
        if logger.parent_with_id:
            command_log.parent_log = '{}|{}'.format(logger.parent_with_id.name, logger.parent_with_id.id)
        logger.backend_logs.elasticsearch = command_log
        command_log.save()

    def command_output_updated(self, logger):
        command_log = logger.backend_logs.elasticsearch
        command_log.update(
            slug=logger.slug,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def command_finished(self, logger):
        command_log = logger.backend_logs.elasticsearch
        command_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CommandState.SUCCEEDED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def command_error(self, logger):
        command_log = logger.backend_logs.elasticsearch
        command_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CommandState.FAILED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def celery_task_invocation_started(self, logger):
        related_objects = [
            get_key_from_object(related_object) for related_object in logger.related_objects
        ]
        celery_task_invocation_log = CeleryTaskInvocationLog(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskInvocationLogState.WAITING,
            related_objects=related_objects,
            **logger.data
        )
        celery_task_invocation_log.meta.id = logger.id
        if logger.parent_with_id:
            celery_task_invocation_log.parent_log = '{}|{}'.format(logger.parent_with_id.name, logger.parent_with_id.id)
        logger.backend_logs.elasticsearch = celery_task_invocation_log
        celery_task_invocation_log.save()

    def celery_task_invocation_triggered(self, logger):
        celery_task_invocation_log = logger.backend_logs.elasticsearch
        celery_task_invocation_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskInvocationLogState.TRIGGERED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def celery_task_invocation_ignored(self, logger):
        celery_task_invocation_log = logger.backend_logs.elasticsearch
        celery_task_invocation_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskInvocationLogState.IGNORED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def celery_task_invocation_timeout(self, logger):
        celery_task_invocation_log = logger.backend_logs.elasticsearch
        celery_task_invocation_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskInvocationLogState.TIMEOUT,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            retry_on_conflict=1,
            **logger.data
        )

    def celery_task_invocation_expired(self, logger):
        celery_task_invocation_log = CeleryTaskInvocationLog.get(id=logger.id)
        celery_task_invocation_log.update(
            slug=logger.slug,
            state=CeleryTaskInvocationLogState.EXPIRED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )
        if celery_task_invocation_log.celery_task_id:
            CeleryTaskRunLog._index.refresh()
            celery_task_run_log_qs = CeleryTaskRunLog.search().filter(
                Q('term', celery_task_id=celery_task_invocation_log.celery_task_id)
                & Q('term', state=CeleryTaskRunLogState.ACTIVE.name)
            )
            for celery_task_run in celery_task_run_log_qs:
                celery_task_run.update(
                    state=CeleryTaskRunLogState.EXPIRED,
                    stop=logger.data['stop']
                )

    def celery_task_run_started(self, logger):
        related_objects = [
            get_key_from_object(related_object) for related_object in logger.related_objects
        ]
        celery_task_run_log = CeleryTaskRunLog(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskRunLogState.ACTIVE,
            related_objects=related_objects,
            **logger.data
        )
        celery_task_run_log.meta.id = logger.id
        if logger.parent_with_id:
            celery_task_run_log.parent_log = '{}|{}'.format(logger.parent_with_id.name, logger.parent_with_id.id)
        logger.backend_logs.elasticsearch = celery_task_run_log
        celery_task_run_log.save()

    def celery_task_run_succeeded(self, logger):
        celery_task_run_log = logger.backend_logs.elasticsearch
        celery_task_run_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskRunLogState.SUCCEEDED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

        CeleryTaskInvocationLog._index.refresh()
        celery_task_invocations_qs = CeleryTaskInvocationLog.search().filter(
            'term', celery_task_id=celery_task_run_log.celery_task_id
        ).query(
            Q('term', state=CeleryTaskInvocationLogState.WAITING.name)
            | Q('term', state=CeleryTaskInvocationLogState.TRIGGERED.name)
            | Q('term', state=CeleryTaskInvocationLogState.ACTIVE.name)
        )
        for celery_task_invocation in celery_task_invocations_qs:
            try:
                celery_task_invocation.update(
                    state=CeleryTaskInvocationLogState.SUCCEEDED,
                    stop=logger.data['stop'],
                    refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                    update_only_changed_fields=True
                )
            except ConflictError:
                # conflict errors are ignored, celery task invocation was changed with another process
                pass

    def celery_task_run_failed(self, logger):
        celery_task_run_log = logger.backend_logs.elasticsearch
        celery_task_run_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskRunLogState.FAILED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

        CeleryTaskInvocationLog._index.refresh()
        celery_task_invocations_qs = CeleryTaskInvocationLog.search().filter(
            'term', celery_task_id=celery_task_run_log.celery_task_id
        ).query(
            Q('term', state=CeleryTaskInvocationLogState.WAITING.name)
            | Q('term', state=CeleryTaskInvocationLogState.TRIGGERED.name)
            | Q('term', state=CeleryTaskInvocationLogState.ACTIVE.name)
        )
        for celery_task_invocation in celery_task_invocations_qs:
            try:
                celery_task_invocation.update(
                    state=CeleryTaskInvocationLogState.FAILED,
                    refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                    stop=logger.data['stop'],
                    update_only_changed_fields=True
                )
            except ConflictError:
                # conflict errors are ignored, celery task invocation was changed with another process
                pass

    def celery_task_run_retried(self, logger):
        celery_task_run_log = logger.backend_logs.elasticsearch
        celery_task_run_log.update(
            slug=logger.slug,
            extra_data=logger.extra_data,
            state=CeleryTaskRunLogState.RETRIED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )

    def celery_task_run_output_updated(self, logger):
        celery_task_run_log = logger.backend_logs.elasticsearch
        celery_task_run_log.update(
            slug=logger.slug,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data,
        )

    def set_stale_celery_task_log_state(self):
        processsing_stale_tasks = CeleryTaskInvocationLog.search().filter(
            Q('range', stale_at={'lt': now()}) & Q(
                Q('term', state=CeleryTaskInvocationLogState.WAITING.name)
                | Q('term', state=CeleryTaskInvocationLogState.TRIGGERED.name)
                | Q('term', state=CeleryTaskInvocationLogState.ACTIVE.name)
            )
        )

        for task in processsing_stale_tasks:
            task_last_run = task.last_run
            if task_last_run and task_last_run.state == CeleryTaskRunLogState.SUCCEEDED:
                task.update(
                    state=CeleryTaskInvocationLogState.SUCCEEDED,
                    stop=task_last_run.stop,
                    update_only_changed_fields=True
                )
            elif task_last_run and task_last_run.state == CeleryTaskRunLogState.FAILED:
                task.update(
                    state=CeleryTaskInvocationLogState.FAILED,
                    stop=task_last_run.stop,
                    update_only_changed_fields=True
                )
            else:
                try:
                    current_app.tasks[task.name].expire_invocation(
                        task.id,
                        json.loads(task.task_args),
                        json.loads(task.task_kwargs),
                        dict(
                            start=task.start,
                            celery_task_id=task.celery_task_id,
                            stop=now(),
                            name=task.name,
                            queue_name=task.queue_name,
                        )
                    )
                except NotRegistered:
                    task.update(
                        state=CeleryTaskInvocationLogState.EXPIRED,
                        stop=now(),
                        update_only_changed_fields=True
                    )

    def clean_logs(self, type, timestamp, backup_path, stdout):
        qs = get_log_model_from_logger_name(type).search().filter(Q('range', start={'lt': timestamp}))
        stdout.write(f'  Removing "{qs.count()}" logs')
        qs.delete()



# Generated by Django 3.1 on 2021-03-26 17:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security', '0002_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='celerytasklog',
            name='triggred_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='trigger time'),
        ),
    ]



# Generated by Django 3.1.13 on 2021-09-06 14:47

import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import enumfields.fields
import security.enums


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CeleryTaskInvocationLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False, verbose_name='log ID')),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='response headers')),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug')),
                ('start', models.DateTimeField(editable=False, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='stop')),
                ('time', models.FloatField(blank=True, null=True, verbose_name='response time')),
                ('celery_task_id', models.UUIDField(blank=True, db_index=True, null=True,
                                                    verbose_name='invocation ID')),
                ('name', models.CharField(db_index=True, max_length=250, verbose_name='task name')),
                ('queue_name', models.CharField(blank=True, max_length=250, null=True, verbose_name='queue name')),
                ('applied_at', models.DateTimeField(verbose_name='applied at')),
                ('triggered_at', models.DateTimeField(blank=True, null=True, verbose_name='triggered at')),
                ('is_unique', models.BooleanField(verbose_name='is unique')),
                ('is_async', models.BooleanField(verbose_name='is async')),
                ('is_duplicate', models.BooleanField(blank=True, null=True, verbose_name='is duplicate')),
                ('is_on_commit', models.BooleanField(verbose_name='is on commit')),
                ('input', models.TextField(blank=True, editable=False, null=True, verbose_name='input')),
                ('task_args',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True, verbose_name='task args')),
                ('task_kwargs',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True, verbose_name='task kwargs')),
                ('estimated_time_of_first_arrival',
                 models.DateTimeField(blank=True, null=True, verbose_name='estimated time of first arrival')),
                ('expires_at', models.DateTimeField(blank=True, null=True, verbose_name='time of expiration')),
                ('stale_at', models.DateTimeField(blank=True, null=True, verbose_name='stale task time')),
                ('state', enumfields.fields.NumEnumField(db_index=True, default=1,
                                                         enum=security.enums.CeleryTaskInvocationLogState,
                                                         verbose_name='state')),
            ],
            options={
                'verbose_name': 'celery task',
                'verbose_name_plural': 'celery tasks',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskRunLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False, verbose_name='log ID')),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='response headers')),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug')),
                ('start', models.DateTimeField(editable=False, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='stop')),
                ('time', models.FloatField(blank=True, null=True, verbose_name='response time')),
                ('celery_task_id', models.UUIDField(blank=True, db_index=True, null=True,
                                                    verbose_name='invocation ID')),
                ('name', models.CharField(db_index=True, max_length=250, verbose_name='task name')),
                ('queue_name', models.CharField(blank=True, max_length=250, null=True, verbose_name='queue name')),
                ('input', models.TextField(blank=True, editable=False, null=True, verbose_name='input')),
                ('task_args',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True, verbose_name='task args')),
                ('task_kwargs',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True, verbose_name='task kwargs')),
                ('state',
                 enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.CeleryTaskRunLogState,
                                                verbose_name='state')),
                ('result',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True, verbose_name='result')),
                ('error_message', models.TextField(blank=True, null=True, verbose_name='error message')),
                ('output', models.TextField(blank=True, editable=False, null=True, verbose_name='output')),
                ('retries', models.PositiveSmallIntegerField(default=0, verbose_name='retries')),
                ('estimated_time_of_next_retry',
                 models.DateTimeField(blank=True, null=True, verbose_name='estimated time of arrival')),
            ],
            options={
                'verbose_name': 'celery task run',
                'verbose_name_plural': 'celery tasks run',
                'ordering': ('created_at',),
            },
        ),
        migrations.CreateModel(
            name='CommandLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False, verbose_name='log ID')),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='response headers')),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug')),
                ('start', models.DateTimeField(editable=False, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='stop')),
                ('time', models.FloatField(blank=True, null=True, verbose_name='response time')),
                ('name', models.CharField(db_index=True, editable=False, max_length=250, verbose_name='name')),
                ('input', models.TextField(editable=False, verbose_name='input')),
                ('is_executed_from_command_line',
                 models.BooleanField(default=False, editable=False, verbose_name='is executed from command line')),
                ('output', models.TextField(blank=True, editable=False, null=True, verbose_name='output')),
                ('state', enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.CommandState,
                                                         verbose_name='state')),
                ('error_message', models.TextField(blank=True, editable=False, null=True,
                                                   verbose_name='error message')),
            ],
            options={
                'verbose_name': 'command log',
                'verbose_name_plural': 'command logs',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='InputRequestLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False, verbose_name='log ID')),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='response headers')),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug')),
                ('start', models.DateTimeField(editable=False, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='stop')),
                ('time', models.FloatField(blank=True, null=True, verbose_name='response time')),
                ('host', models.CharField(db_index=True, max_length=255, verbose_name='host')),
                ('method', models.SlugField(max_length=7, verbose_name='method')),
                ('path', models.CharField(blank=True, db_index=True, max_length=2000, verbose_name='URL path')),
                ('queries',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='queries')),
                ('is_secure', models.BooleanField(default=False, verbose_name='HTTPS connection')),
                ('request_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='request headers')),
                ('request_body', models.TextField(null=True, blank=True, verbose_name='request body')),
                ('response_code', models.PositiveSmallIntegerField(blank=True, null=True,
                                                                   verbose_name='response code')),
                ('response_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='response headers')),
                ('response_body', models.TextField(blank=True, null=True, verbose_name='response body')),
                ('state',
                 enumfields.fields.NumEnumField(default=0, enum=security.enums.RequestLogState, verbose_name='state')),
                ('error_message', models.TextField(blank=True, null=True, verbose_name='error description')),
                ('user_id', models.TextField(blank=True, db_index=True, null=True, verbose_name='user ID')),
                ('ip', models.GenericIPAddressField(db_index=True, verbose_name='IP address')),
                ('view_slug',
                 models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='view slug')),
            ],
            options={
                'verbose_name': 'input logged request',
                'verbose_name_plural': 'input logged requests',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OutputRequestLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False, verbose_name='log ID')),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='response headers')),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True, verbose_name='slug')),
                ('start', models.DateTimeField(editable=False, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='stop')),
                ('time', models.FloatField(blank=True, null=True, verbose_name='response time')),
                ('host', models.CharField(db_index=True, max_length=255, verbose_name='host')),
                ('method', models.SlugField(max_length=7, verbose_name='method')),
                ('path', models.CharField(blank=True, db_index=True, max_length=2000, verbose_name='URL path')),
                ('queries',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='queries')),
                ('is_secure', models.BooleanField(default=False, verbose_name='HTTPS connection')),
                ('request_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='request headers')),
                ('request_body', models.TextField(null=True, blank=True, verbose_name='request body')),
                ('response_code', models.PositiveSmallIntegerField(blank=True, null=True,
                                                                   verbose_name='response code')),
                ('response_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True,
                                  verbose_name='response headers')),
                ('response_body', models.TextField(blank=True, null=True, verbose_name='response body')),
                ('state',
                 enumfields.fields.NumEnumField(default=0, enum=security.enums.RequestLogState, verbose_name='state')),
                ('error_message', models.TextField(blank=True, null=True, verbose_name='error description')),
            ],
            options={
                'verbose_name': 'output logged request',
                'verbose_name_plural': 'output logged requests',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OutputRequestLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('output_request_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.outputrequestlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('output_request_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='InputRequestLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('input_request_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.inputrequestlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('input_request_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CommandLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('command_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.commandlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('command_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskRunLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('celery_task_run_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.celerytaskrunlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('celery_task_run_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskInvocationLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('celery_task_invocation_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects',
                                   to='security_backends_sql.celerytaskinvocationlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('celery_task_invocation_log', 'object_ct_id', 'object_id')},
            },
        ),
    ]



import logging

from security.backends.writer import BaseBackendWriter


input_request_logger = logging.getLogger('security.input_request')
output_request_logger = logging.getLogger('security.output_request')
command_logger = logging.getLogger('security.command')
celery_logger = logging.getLogger('security.celery')


class LoggingBackendWriter(BaseBackendWriter):

    def input_request_started(self, logger):
        input_request_logger.info(
            'Input request "%(id)s" to "%(host)s" with path "%(path)s" was started',
            dict(
                id=logger.id,
                host=logger.data['host'],
                path=logger.data['path'],
            ),
            extra=dict(
                input_request_id=logger.id,
                input_request_host=logger.data['host'],
                input_request_path=logger.data['path'],
                input_request_method=logger.data['method'],
                input_request_view_slug=logger.data['view_slug'],
                input_request_slug=logger.slug,
                input_request_is_secure=logger.data['is_secure'],
                input_request_ip=logger.data['ip'],
            )
        )

    def input_request_finished(self, logger):
        input_request_logger.info(
            'Input request "%(id)s" to "%(host)s" with path "%(path)s" was finished',
            dict(
                id=logger.id,
                host=logger.data['host'],
                path=logger.data['path'],
            ),
            extra=dict(
                input_request_id=logger.id,
                input_request_host=logger.data['host'],
                input_request_path=logger.data['path'],
                input_request_method=logger.data['method'],
                input_request_view_slug=logger.data['view_slug'],
                input_request_slug=logger.slug,
                input_request_is_secure=logger.data['is_secure'],
                input_request_ip=logger.data['ip'],
                input_request_response_code=logger.data['response_code'],
                input_request_start=logger.data['start'],
                input_request_stop=logger.data['stop'],
                input_request_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
            )
        )

    def input_request_error(self, logger):
        input_request_logger.error(
            'Input request "%(id)s" to "%(host)s" with path "%(path)s" failed',
            dict(
                id=logger.id,
                host=logger.data['host'],
                path=logger.data['path'],
            ),
            extra=dict(
                input_request_id=logger.id,
                input_request_host=logger.data['host'],
                input_request_path=logger.data['path'],
                input_request_method=logger.data['method'],
                input_request_view_slug=logger.data['view_slug'],
                input_request_slug=logger.slug,
                input_request_is_secure=logger.data['is_secure'],
                input_request_ip=logger.data['ip'],
                input_request_start=logger.data['start'],
            )
        )

    def output_request_started(self, logger):
        output_request_logger.info(
            'Output request "%(id)s" to "%(host)s" with path "%(path)s" was started',
            dict(
                id=logger.id,
                host=logger.data['host'],
                path=logger.data['path'],
            ),
            extra=dict(
                output_request_id=logger.id,
                output_request_host=logger.data['host'],
                output_request_path=logger.data['path'],
                output_request_method=logger.data['method'],
                output_request_slug=logger.slug,
                output_request_is_secure=logger.data['is_secure'],
            )
        )

    def output_request_finished(self, logger):
        output_request_logger.info(
            'Output request "%(id)s" to "%(host)s" with path "%(path)s" was successful',
            dict(
                id=logger.id,
                host=logger.data['host'],
                path=logger.data['path'],
            ),
            extra=dict(
                output_request_id=logger.id,
                output_request_host=logger.data['host'],
                output_request_path=logger.data['path'],
                output_request_method=logger.data['method'],
                output_request_slug=logger.slug,
                output_request_is_secure=logger.data['is_secure'],
                output_request_start=logger.data['start'],
                output_request_stop=logger.data['stop'],
                output_request_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
            )
        )

    def output_request_error(self, logger):
        output_request_logger.error(
            'Output request "%(id)s" to "%(host)s" with path "%(path)s" failed',
            dict(
                id=logger.id,
                host=logger.data['host'],
                path=logger.data['path'],
            ),
            extra=dict(
                output_request_id=logger.id,
                output_request_host=logger.data['host'],
                output_request_path=logger.data['path'],
                output_request_method=logger.data['method'],
                output_request_slug=logger.slug,
                output_request_is_secure=logger.data['is_secure'],
                output_request_start=logger.data['start'],
                output_request_stop=logger.data['stop'],
                output_request_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
            )
        )

    def command_started(self, logger):
        command_logger.info(
            'Command "%(id)s" with name "%(name)s" was started',
            dict(
                id=logger.id,
                name=logger.data['name'],
            ),
            extra=dict(
                command_id=logger.id,
                command_name=logger.data['name'],
                command_is_executed_from_command_line=logger.data['is_executed_from_command_line'],
                command_start=logger.data['start'],
            )
        )

    def command_output_updated(self, logger):
        """Output is not stored into log"""

    def command_finished(self, logger):
        command_logger.info(
            'Command "%(id)s" with name "%(name)s" was successful',
            dict(
                id=logger.id,
                name=logger.data['name'],
            ),
            extra=dict(
                command_id=logger.id,
                command_name=logger.data['name'],
                command_is_executed_from_command_line=logger.data['is_executed_from_command_line'],
                command_start=logger.data['start'],
                command_stop=logger.data['stop'],
                command_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
                command_is_error=False
            )
        )

    def command_error(self, logger):
        command_logger.error(
            'Command "%(id)s" with name "%(name)s" failed',
            dict(
                id=logger.id,
                name=logger.data['name'],
            ),
            extra=dict(
                command_id=logger.id,
                command_name=logger.data['name'],
                command_is_executed_from_command_line=logger.data['is_executed_from_command_line'],
                command_start=logger.data['start'],
                command_stop=logger.data['stop'],
                command_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
                command_is_error=True
            )
        )

    def celery_task_invocation_started(self, logger):
        """Invocation started means nothing for logging backend"""

    def celery_task_invocation_triggered(self, logger):
        celery_logger.info(
            'Celery task invocation "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was invoked',
            dict(
                id=logger.id,
                celery_task_id=logger.data['celery_task_id'],
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.data['celery_task_id'],
                celery_task_invocation_name=logger.data['name'],
                celery_task_invocation_queue=logger.data['queue_name'],
                celery_task_invocation_start=logger.data['start'],
                celery_task_invocation_state='TRIGGERED',
            ),
        )

    def celery_task_invocation_ignored(self, logger):
        celery_logger.info(
            'Celery task invocation "%(id)s" with name "%(name)s" was ignored',
            dict(
                id=logger.id,
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_name=logger.data['name'],
                celery_task_invocation_queue=logger.data['queue_name'],
                celery_task_invocation_start=logger.data['start'],
                celery_task_invocation_stop=logger.data['stop'],
                celery_task_invocation_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
                celery_task_invocation_state='IGNORED',
            ),
        )

    def celery_task_invocation_timeout(self, logger):
        celery_logger.warning(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" caused a response timeout',
            dict(
                id=logger.id,
                celery_task_id=logger.data['celery_task_id'],
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.data['celery_task_id'],
                celery_task_invocation_name=logger.data['name'],
                celery_task_invocation_queue=logger.data['queue_name'],
                celery_task_invocation_start=logger.data['start'],
                celery_task_invocation_stop=logger.data['stop'],
                celery_task_invocation_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
                celery_task_invocation_state='TIMEOUT',
            ),
        )

    def celery_task_invocation_expired(self, logger):
        celery_logger.error(
            'Celery task invocation "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was expired',
            dict(
                id=logger.id,
                celery_task_id=logger.data['celery_task_id'] or '',
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.data['celery_task_id'] or '',
                celery_task_invocation_name=logger.data['name'],
                celery_task_invocation_queue=logger.data['queue_name'],
                celery_task_invocation_start=logger.data['start'],
                celery_task_invocation_stop=logger.data['stop'],
                celery_task_invocation_time=(logger.data['stop'] - logger.data['start']).total_seconds(),
                task_state='EXPIRED',
            ),
        )

    def celery_task_run_started(self, logger):
        celery_logger.info(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was started',
            dict(
                id=logger.id,
                celery_task_id=logger.data['celery_task_id'],
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.data['name'],
                celery_task_queue=logger.data['queue_name'],
                celery_task_tart=logger.data['start'],
                celery_task_invocation_state='ACTIVE',
                celery_task_attempt=logger.data['retries'],
            ),
        )

    def celery_task_run_succeeded(self, logger):
        celery_logger.info(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was successful',
            dict(
                id=logger.id,
                celery_task_id=logger.data['celery_task_id'],
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.data['name'],
                celery_task_queue=logger.data['queue_name'],
                celery_task_tart=logger.data['start'],
                celery_task_invocation_state='SUCCEEDED',
                celery_task_attempt=logger.data['retries'],
                celery_task_stop=logger.data['stop'],
                celery_task_time=logger.data['stop'] - logger.data['start']
            )
        )

    def celery_task_run_failed(self, logger):
        celery_logger.error(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" failed',
            dict(
                id=logger.id,
                celery_task_id=logger.data['celery_task_id'],
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.data['name'],
                celery_task_queue=logger.data['queue_name'],
                celery_task_tart=logger.data['start'],
                celery_task_invocation_state='FAILED',
                celery_task_attempt=logger.data['retries'],
                celery_task_stop=logger.data['stop'],
                celery_task_time=logger.data['stop'] - logger.data['start']
            )
        )

    def celery_task_run_retried(self, logger):
        celery_logger.warning(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was repeated',
            dict(
                id=logger.id,
                celery_task_id=logger.data['celery_task_id'],
                name=logger.data['name'],
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.data['name'],
                celery_task_queue=logger.data['queue_name'],
                celery_task_tart=logger.data['start'],
                celery_task_invocation_state='RETRIED',
                celery_task_attempt=logger.data['retries'],
                celery_task_stop=logger.data['stop'],
                celery_task_time=logger.data['stop'] - logger.data['start']
            )
        )

    def celery_task_run_output_updated(self, logger):
        """Output is not stored into log"""



import os
import json
import gzip
import math

from io import TextIOWrapper

from datetime import datetime, time

from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.timezone import utc
from django.contrib.contenttypes.models import ContentType
from django.utils.timezone import now

from celery import current_app
from celery.exceptions import NotRegistered

from chamber.shortcuts import change_and_save

from security.backends.writer import BaseBackendWriter
from security.config import settings
from security.enums import RequestLogState, CeleryTaskInvocationLogState, CeleryTaskRunLogState, CommandState

from .models import (
    InputRequestLog, OutputRequestLog, CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog,
    get_log_model_from_logger_name, get_response_state
)


def lazy_serialize_qs_without_pk(qs):
    def generator():
        for obj in qs.iterator():
            data = serializers.serialize('python', [obj])[0]

            del data['pk']
            yield data

    class StreamList(list):
        def __iter__(self):
            return generator()

        # according to the comment below
        def __len__(self):
            return 1

    return StreamList()


def get_querysets_by_batch(qs, batch):
    steps = math.ceil(qs.count() / batch)
    for _ in range(steps):
        yield qs[:batch]


class SQLBackendWriter(BaseBackendWriter):

    def input_request_started(self, logger):
        input_request_log = InputRequestLog(
            id=logger.id,
            slug=logger.slug,
            state=RequestLogState.INCOMPLETE,
            extra_data=logger.extra_data,
            **logger.data
        )
        logger.backend_logs.sql = input_request_log
        input_request_log.save()
        input_request_log.related_objects.add(*logger.related_objects)
        if logger.parent_with_id:
            input_request_log.related_objects.create(
                object_ct_id=ContentType.objects.get_for_model(
                    _get_log_model_from_logger_name(logger.parent_with_id.name)
                ).pk,
                object_id=logger.parent_with_id.id
            )

    def input_request_finished(self, logger):
        input_request_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=get_response_state(logger.data['response_code']),
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        input_request_log.related_objects.add(*logger.related_objects)

    def input_request_error(self, logger):
        input_request_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        input_request_log.related_objects.add(*logger.related_objects)

    def output_request_started(self, logger):
        output_request_log = OutputRequestLog(
            id=logger.id,
            state=RequestLogState.INCOMPLETE,
            extra_data=logger.extra_data,
            **logger.data
        )
        logger.backend_logs.sql = output_request_log
        output_request_log.save()
        output_request_log.related_objects.add(*logger.related_objects)
        if logger.parent_with_id:
            output_request_log.related_objects.create(
                object_ct_id=ContentType.objects.get_for_model(
                    _get_log_model_from_logger_name(logger.parent_with_id.name)
                ).pk,
                object_id=logger.parent_with_id.id
            )

    def output_request_finished(self, logger):
        output_request_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=get_response_state(logger.data['response_code']),
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        output_request_log.related_objects.add(*logger.related_objects)

    def output_request_error(self, logger):
        output_request_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=RequestLogState.ERROR,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        output_request_log.related_objects.add(*logger.related_objects)

    def command_started(self, logger):
        command_log = CommandLog(
            id=logger.id,
            slug=logger.slug,
            state=CommandState.ACTIVE,
            extra_data=logger.extra_data,
            **logger.data
        )
        logger.backend_logs.sql = command_log
        command_log.save()
        command_log.related_objects.add(*logger.related_objects)
        if logger.parent_with_id:
            command_log.related_objects.create(
                object_ct_id=ContentType.objects.get_for_model(
                    _get_log_model_from_logger_name(logger.parent_with_id.name)
                ).pk,
                object_id=logger.parent_with_id.id
            )

    def command_output_updated(self, logger):
        change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            update_only_changed_fields=True,
            **logger.data
        )

    def command_finished(self, logger):
        command_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CommandState.SUCCEEDED,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        command_log.related_objects.add(*logger.related_objects)

    def command_error(self, logger):
        command_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CommandState.FAILED,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        command_log.related_objects.add(*logger.related_objects)

    def celery_task_invocation_started(self, logger):
        celery_task_invocation_log = CeleryTaskInvocationLog(
            id=logger.id,
            slug=logger.slug,
            state=CeleryTaskInvocationLogState.WAITING,
            extra_data=logger.extra_data,
            **logger.data
        )
        logger.backend_logs.sql = celery_task_invocation_log
        celery_task_invocation_log.save()
        celery_task_invocation_log.related_objects.add(*logger.related_objects)
        if logger.parent_with_id:
            celery_task_invocation_log.related_objects.create(
                object_ct_id=ContentType.objects.get_for_model(
                    _get_log_model_from_logger_name(logger.parent_with_id.name)
                ).pk,
                object_id=logger.parent_with_id.id
            )

    def celery_task_invocation_triggered(self, logger):
        celery_task_invocation_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CeleryTaskInvocationLogState.TRIGGERED,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        celery_task_invocation_log.related_objects.add(*logger.related_objects)

    def celery_task_invocation_ignored(self, logger):
        celery_task_invocation_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CeleryTaskInvocationLogState.IGNORED,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        celery_task_invocation_log.related_objects.add(*logger.related_objects)

    def celery_task_invocation_timeout(self, logger):
        celery_task_invocation_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CeleryTaskInvocationLogState.TIMEOUT,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        celery_task_invocation_log.related_objects.add(*logger.related_objects)

    def celery_task_invocation_expired(self, logger):
        celery_task_invocation_log = CeleryTaskInvocationLog.objects.update_or_create(
            id=logger.id,
            defaults=dict(
                slug=logger.slug,
                state=CeleryTaskInvocationLogState.EXPIRED,
                extra_data=logger.extra_data,
                **logger.data
            )
        )[0]
        celery_task_invocation_log.runs.filter(
            state=CeleryTaskRunLogState.ACTIVE
        ).change_and_save(
            state=CeleryTaskRunLogState.EXPIRED,
            stop=logger.data['stop']
        )
        celery_task_invocation_log.related_objects.add(*logger.related_objects)

    def celery_task_run_started(self, logger):
        celery_task_run_log = CeleryTaskRunLog(
            id=logger.id,
            slug=logger.slug,
            state=CeleryTaskRunLogState.ACTIVE,
            extra_data=logger.extra_data,
            **logger.data
        )
        logger.backend_logs.sql = celery_task_run_log
        celery_task_run_log.save()
        celery_task_run_log.related_objects.add(*logger.related_objects)
        if logger.parent_with_id:
            celery_task_run_log.related_objects.create(
                object_ct_id=ContentType.objects.get_for_model(
                    _get_log_model_from_logger_name(logger.parent_with_id.name)
                ).pk,
                object_id=logger.parent_with_id.id
            )

    def celery_task_run_succeeded(self, logger):
        celery_task_run_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CeleryTaskRunLogState.SUCCEEDED,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        celery_task_run_log.get_task_invocation_logs().filter(state__in={
            CeleryTaskInvocationLogState.WAITING,
            CeleryTaskInvocationLogState.TRIGGERED,
            CeleryTaskInvocationLogState.ACTIVE
        }).change_and_save(
            state=CeleryTaskInvocationLogState.SUCCEEDED,
            stop=logger.data['stop']
        )
        celery_task_run_log.related_objects.add(*logger.related_objects)

    def celery_task_run_failed(self, logger):
        celery_task_run_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CeleryTaskRunLogState.FAILED,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        celery_task_run_log.get_task_invocation_logs().filter(state__in={
            CeleryTaskInvocationLogState.WAITING,
            CeleryTaskInvocationLogState.TRIGGERED,
            CeleryTaskInvocationLogState.ACTIVE
        }).change_and_save(
            state=CeleryTaskInvocationLogState.FAILED,
            stop=logger.data['stop']
        )
        celery_task_run_log.related_objects.add(*logger.related_objects)

    def celery_task_run_retried(self, logger):
        celery_task_run_log = change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            state=CeleryTaskRunLogState.RETRIED,
            extra_data=logger.extra_data,
            update_only_changed_fields=True,
            **logger.data
        )
        celery_task_run_log.related_objects.add(*logger.related_objects)

    def celery_task_run_output_updated(self, logger):
        change_and_save(
            logger.backend_logs.sql,
            slug=logger.slug,
            update_only_changed_fields=True,
            **logger.data
        )

    def set_stale_celery_task_log_state(self):
        for task in CeleryTaskInvocationLog.objects.filter_processing(stale_at__lt=now()).iterator():
            task_last_run = task.last_run
            if task_last_run and task_last_run.state == CeleryTaskRunLogState.SUCCEEDED:
                change_and_save(
                    task,
                    state=CeleryTaskInvocationLogState.SUCCEEDED,
                    stop=task_last_run.stop,
                    update_only_changed_fields=True
                )
            elif task_last_run and task_last_run.state == CeleryTaskRunLogState.FAILED:
                change_and_save(
                    task,
                    state=CeleryTaskInvocationLogState.FAILED,
                    stop=task_last_run.stop,
                    update_only_changed_fields=True
                )
            else:
                try:
                    current_app.tasks[task.name].expire_invocation(
                        task.id,
                        task.task_args,
                        task.task_kwargs,
                        dict(
                            start=task.start,
                            celery_task_id=task.celery_task_id,
                            stop=now(),
                            name=task.name,
                            queue_name=task.queue_name,
                        )
                    )
                except NotRegistered:
                    change_and_save(
                        task,
                        state=CeleryTaskInvocationLogState.EXPIRED,
                        stop=now(),
                        update_only_changed_fields=True
                    )

    def clean_logs(self, type, timestamp, backup_path, stdout):
        qs = get_log_model_from_logger_name(type).objects.filter(stop__lte=timestamp).order_by('stop')
        for timestamp in qs.datetimes('start', 'day', tzinfo=utc):
            min_timestamp = datetime.combine(timestamp, time.min).replace(tzinfo=utc)
            max_timestamp = datetime.combine(timestamp, time.max).replace(tzinfo=utc)
            qs_filtered_by_day = qs.filter(start__range=(min_timestamp, max_timestamp))

            for qs_batch in get_querysets_by_batch(qs_filtered_by_day, settings.PURGE_LOG_BACKUP_BATCH):
                stdout.write(
                    2 * ' ' + 'Cleaning logs for date {} ({})'.format(
                        timestamp.date(), qs_batch.count()
                    )
                )
                if backup_path:
                    log_file_name = os.path.join(backup_path, str(timestamp.date()))

                    if storage.exists('{}.json.gz'.format(log_file_name)):
                        i = 1
                        while storage.exists('{}({}).json.gz'.format(log_file_name, i)):
                            i += 1
                        log_file_name = '{}({})'.format(log_file_name, i)

                    stdout.write(4 * ' ' + 'generating backup file: {}.json.gz'.format(log_file_name))

                    with storage.open('{}.json.gz'.format(log_file_name), 'wb') as f:
                        with TextIOWrapper(gzip.GzipFile(filename='{}.json'.format(log_file_name),
                                                         fileobj=f, mode='wb')) as gzf:
                            json.dump(
                                lazy_serialize_qs_without_pk(qs_batch), gzf, cls=DjangoJSONEncoder, indent=5
                            )
                stdout.write(4 * ' ' + 'deleting logs')

                for qs_batch_to_delete in get_querysets_by_batch(qs_batch, settings.PURGE_LOG_DELETE_BATCH):
                    qs_filtered_by_day.filter(pk__in=qs_batch_to_delete).delete()



# Generated by Django 3.1.7 on 2021-04-01 07:58

import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import enumfields.fields
import security.old.enums


class Migration(migrations.Migration):
    dependencies = [
        ('security', '0003_migration'),
    ]

    operations = [
        migrations.CreateModel(
            name='CeleryTaskInvocationLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('invocation_id',
                 models.CharField(db_index=True, max_length=250, unique=True, verbose_name='invocation ID')),
                ('celery_task_id',
                 models.CharField(blank=True, db_index=True, max_length=250, null=True, verbose_name='celery ID')),
                ('name', models.CharField(db_index=True, max_length=250, verbose_name='task name')),
                ('queue_name', models.CharField(blank=True, max_length=250, null=True, verbose_name='queue name')),
                ('applied_at', models.DateTimeField(verbose_name='applied at')),
                ('triggered_at', models.DateTimeField(blank=True, null=True, verbose_name='triggered at')),
                ('is_unique', models.BooleanField(verbose_name='is unique')),
                ('is_async', models.BooleanField(verbose_name='is async')),
                ('is_duplicate', models.BooleanField(verbose_name='is duplicate')),
                ('is_on_commit', models.BooleanField(verbose_name='is on commit')),
                ('input', models.TextField(blank=True, editable=False, null=True, verbose_name='input')),
                ('task_args',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True, verbose_name='task args')),
                ('task_kwargs',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True, verbose_name='task kwargs')),
                ('estimated_time_of_first_arrival',
                 models.DateTimeField(blank=True, null=True, verbose_name='estimated time of first arrival')),
                ('expires_at', models.DateTimeField(blank=True, null=True, verbose_name='time of expiration')),
                ('stale_at', models.DateTimeField(blank=True, null=True, verbose_name='stale task time')),
                ('state', enumfields.fields.NumEnumField(db_index=True, default=1,
                                                         enum=security.old.enums.CeleryTaskInvocationLogState,
                                                         verbose_name='state')),
            ],
            options={
                'verbose_name': 'celery task',
                'verbose_name_plural': 'celery tasks',
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskInvocationLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object')),
                ('object_id', models.TextField(db_index=True, verbose_name='ID of the related object')),
                ('celery_task_invocation_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security.celerytaskinvocationlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('celery_task_invocation_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.AlterUniqueTogether(
            name='celerytaskloggenericmanytomanyrelation',
            unique_together=None,
        ),
        migrations.RemoveField(
            model_name='celerytaskloggenericmanytomanyrelation',
            name='celery_task_log',
        ),
        migrations.AddField(
            model_name='celerytaskrunlog',
            name='queue_name',
            field=models.CharField(blank=True, max_length=250, null=True, verbose_name='queue name'),
        ),
        migrations.DeleteModel(
            name='CeleryTaskLog',
        ),
        migrations.DeleteModel(
            name='CeleryTaskLogGenericManyToManyRelation',
        ),
    ]
